/* Copyright (c) 2016 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */
#include <stdbool.h>
#include <stdint.h>

#include "nrf.h"
#include "app_test_runner.h"
#include "nrf_drv_timer.h"
#include "nrf_drv_clock.h"
#include "FreeRTOS.h"
#include "task.h"
#ifdef SOFTDEVICE_PRESENT
#include "softdevice_handler.h"
#endif

#define TEST_MODE AUTOMATIC_TEST_MODE
#define TEST_NAME "FreeRTOS tickles idle test"

/**
 * @brief Expected tick value when timer handler it executed
 */
static TickType_t m_expected_tickcount_on_timer;

/**
 * @brief The instance of hardware timer for test
 */
static const nrf_drv_timer_t timer = NRF_DRV_TIMER_INSTANCE(1);

#ifdef SOFTDEVICE_PRESENT
static nrf_clock_lf_cfg_t m_sd_clock_config = NRF_CLOCK_LFCLKSRC;
static uint32_t           m_sd_dummy_evt_buff;
#endif

/**
 * @brief Timer configuration
 */
static const nrf_drv_timer_config_t timer_config =
{
    .frequency          = NRF_TIMER_FREQ_1MHz,
    .mode               = NRF_TIMER_MODE_TIMER,
    .bit_width          = NRF_TIMER_BIT_WIDTH_16,
    .interrupt_priority = TIMER_DEFAULT_CONFIG_IRQ_PRIORITY,
    .p_context          = NULL

};

void setUp(void)
{
    nrf_drv_clock_hfclk_request(NULL);
    nrf_drv_clock_lfclk_request(NULL); /* Just to be sure, it would be required by OS. */
    while (!(nrf_drv_clock_hfclk_is_running() &&
             nrf_drv_clock_lfclk_is_running()))
    {
        /* Just waiting */
    }
}

void tearDown(void)
{

}

static void timer_handler(nrf_timer_event_t event_type, void * p_context)
{
    (void)p_context;
    TickType_t current = xTaskGetTickCountFromISR();
    nrf_drv_timer_disable(&timer);
    TEST_ASSERT_UINT_WITHIN(2, m_expected_tickcount_on_timer + 1, current);
}


static void timer_single_shot_test(uint32_t ms)
{
    uint32_t ticks = nrf_drv_timer_ms_to_ticks(&timer, ms);
    TEST_ASSERT(ticks < (1U<<16));

    m_expected_tickcount_on_timer = xTaskGetTickCount() + ms;
    nrf_drv_timer_compare(
        &timer,
        NRF_TIMER_CC_CHANNEL0,
        ticks,
        true);

    nrf_drv_timer_enable(&timer);
}

void test_tick_count_from_isr(void)
{
    unsigned int n;

    nrf_drv_timer_init(&timer, &timer_config, &timer_handler);

    /* Test timing just a few times */
    for(n=0; n<16; ++n)
    {
        timer_single_shot_test(20 + n * 2);
        vTaskDelay(200);
    }
}

#ifdef SOFTDEVICE_PRESENT
void test_tick_count_with_sd(void)
{
    ret_code_t ret;
    unsigned int n;

    nrf_drv_timer_init(&timer, &timer_config, &timer_handler);

    ret = softdevice_handler_init(
        &m_sd_clock_config,
        &m_sd_dummy_evt_buff,
        sizeof(m_sd_dummy_evt_buff),
        NULL);\
    TEST_ASSERT_EQUAL(NRF_SUCCESS, ret);

    /* Test timing just a few times */
    for(n=0; n<8; ++n)
    {
        timer_single_shot_test(30 + n * 3);
        vTaskDelay(200);
    }

    ret = softdevice_handler_sd_disable();
    TEST_ASSERT_EQUAL(NRF_SUCCESS, ret);

    /* Test timing just a few times */
    for(n=0; n<8; ++n)
    {
        timer_single_shot_test(20 + n * 4);
        vTaskDelay(80);
    }
}
#endif

case_t m_test_cases[] = {
    TEST_CASE(test_tick_count_from_isr),
#ifdef SOFTDEVICE_PRESENT
    TEST_CASE(test_tick_count_with_sd),
#endif
};
